@@ -188,6 +188,37 @@ FmFileInfo* fm_file_info_new ()
     return fi;
 }
 
+/**
+ * _fm_file_info_set_emblems:
+ * @fi:  A FmFileInfo struct
+ * @inf: A GFileInfo object
+ *
+ * Read icon emblems metadata from the "metadata::emblems" attribute of
+ * a GFileInfo object and make the GIcon store in the FmFileInfo a 
+ * GEmblemedIcon object if the file has emblems.
+ */
+static void _fm_file_info_set_emblems(FmFileInfo* fi, GFileInfo* inf)
+{
+    char** emblem_names = g_file_info_get_attribute_stringv(inf, "metadata::emblems");
+    if(emblem_names)
+    {
+        GIcon* gicon = g_emblemed_icon_new(G_ICON(fi->icon), NULL);
+        char** emblem_name;
+        for(emblem_name = emblem_names; *emblem_name; ++emblem_name)
+        {
+            GIcon* emblem_icon = g_themed_icon_new(*emblem_name);
+            GEmblem* emblem = g_emblem_new(emblem_icon);
+            g_object_unref(emblem_icon);
+            g_emblemed_icon_add_emblem(G_EMBLEMED_ICON(gicon), emblem);
+            g_object_unref(emblem);
+        }
+        /* replace the original GIcon with an GEmblemedIcon */
+        g_object_unref(fi->icon);
+        fi->icon = fm_icon_from_gicon(gicon);
+        g_object_unref(gicon);
+    }
+}
+
 /**
  * fm_file_info_set_from_native_file:
  * @fi:  A FmFileInfo struct
@@ -213,6 +244,9 @@ gboolean _fm_file_info_set_from_native_file(FmFileInfo* fi, const char* path,
     g_return_val_if_fail(fi && fi->path, FALSE);
     if(lstat(path, &st) == 0)
     {
+        GFile* gfile;
+        GFileInfo* inf;
+
         fi->mode = st.st_mode;
         fi->mtime = st.st_mtime;
         fi->atime = st.st_atime;
@@ -410,6 +444,16 @@ _not_desktop_entry:
         if(!fi->icon)
             fi->icon = g_object_ref(fm_mime_type_get_icon(fi->mime_type));
 
+        gfile = g_file_new_for_path(path);
+        
+        /* get emblems using gio/gvfs-metadata */
+        inf = g_file_query_info(gfile, "metadata::emblems,standard::icon", G_FILE_QUERY_INFO_NONE, NULL, NULL);
+        if(inf)
+        {
+            _fm_file_info_set_emblems(fi, inf);
+            g_object_unref(inf);
+        }
+
         if (!dname)
             dname = g_filename_display_basename(path);
         _fm_path_set_display_name(fi->path, dname);
@@ -419,17 +463,15 @@ _not_desktop_entry:
         fi->fs_is_ro = FALSE;
         if (S_ISDIR(st.st_mode))
         {
-            GFile *gf = g_file_new_for_path(path);
-            GFileInfo *inf;
-            inf = g_file_query_filesystem_info(gf, G_FILE_ATTRIBUTE_FILESYSTEM_READONLY,
+            inf = g_file_query_filesystem_info(gfile, G_FILE_ATTRIBUTE_FILESYSTEM_READONLY,
                                                NULL, NULL);
             if (inf)
             {
                 fi->fs_is_ro = g_file_info_get_attribute_boolean(inf, G_FILE_ATTRIBUTE_FILESYSTEM_READONLY);
                 g_object_unref(inf);
             }
-            g_object_unref(gf);
         }
+        g_object_unref(gfile);
     }
     else
     {
@@ -653,6 +695,9 @@ _file_is_symlink:
     else
         fi->icon = g_object_ref(fm_mime_type_get_icon(fi->mime_type));
 
+    /* if the file has emblems, add them to the icon */
+    _fm_file_info_set_emblems(fi, inf);
+
     if(fm_path_is_native(fi->path))
     {
         fi->dev = g_file_info_get_attribute_uint32(inf, G_FILE_ATTRIBUTE_UNIX_DEVICE);
@@ -1316,7 +1361,8 @@ gboolean fm_file_info_is_unknown_type(FmFileInfo* fi)
 /* full path of the file is required by this function */
 gboolean fm_file_info_is_executable_type(FmFileInfo* fi)
 {
-    if(strncmp(fm_mime_type_get_type(fi->mime_type), "text/", 5) == 0)
+    const char* type = fm_mime_type_get_type(fi->mime_type);
+    if(strncmp(type, "text/", 5) == 0)
     { /* g_content_type_can_be_executable reports text files as executables too */
         /* We don't execute remote files nor files in trash */
         if(fm_path_is_native(fi->path) && (fi->mode & (S_IXOTH|S_IXGRP|S_IXUSR)))
@@ -1335,6 +1381,13 @@ gboolean fm_file_info_is_executable_type(FmFileInfo* fi)
         }
         return FALSE;
     }
+    else if(strcmp(type, "application/x-desktop") == 0)
+    { /* treat desktop entries as executables if
+         they are native and have read permission */
+        if(fm_path_is_native(fi->path) && (fi->mode & (S_IRUSR|S_IRGRP|S_IROTH)))
+            return TRUE;
+        return FALSE;
+    }
     return g_content_type_can_be_executable(fm_mime_type_get_type(fi->mime_type));
 }
 
